#define dbgprintf(a,...) chprintf((BaseSequentialStream*)&SD,a,__VA_ARGS__)
//#define dbgprintf(a,...)

#if 1
void *mp3_malloc(int s) {return chHeapAlloc(NULL, s);}
void mp3_free(void *p) {chHeapFree(p);}
#else
unsigned char buf1[22652];
unsigned char buf2[2568];
unsigned char buf3[4608];

void *mp3_malloc(int s) {void *p; 
                           if (s==22652) p=buf1;
                           else if (s==2567) p=buf2;
                           else if (s==4608) p=buf3;
                           else p=chHeapAlloc(NULL, s); 
                           dbgprintf("Malloc @%x - %d\n\r",p,s); return p;}
void mp3_free(void *p) {dbgprintf("Free @%x\n\r",p); 
                           if ((p!=buf1)&&(p!=buf2)&&(p!=buf3)) chHeapFree(p);}
#endif

void *mp3_calloc(int num,int size)
{
 unsigned char *ptr;
 int i;
 dbgprintf("Calloc %d %d -> ",num,size);
 ptr=mp3_malloc(size*num);
 if (!ptr) return NULL;
 for(i=0;i<num*size;i++) ptr[i]=0;
 return ptr;
}

struct buffer {
  unsigned char *start;
  int eof;
  FIL *file;
  int samplerate;
};

uint16_t wave_buf0[1152*2];
uint16_t wave_buf1[1152*2];
uint16_t playing;

#define MP3BUF 4096

unsigned char mp3buf[MP3BUF];
FIL fIn;
//FIL fOut;
struct buffer buffer;
struct mad_decoder decoder;

/*
 * This is the input callback. The purpose of this callback is to (re)fill
 * the stream buffer which is to be decoded. In this example, an entire file
 * has been mapped into memory, so we just call mad_stream_buffer() with the
 * address and length of the mapping. When this callback is called a second
 * time, we are finished decoding.
 */

enum mad_flow input(void *data,
		    struct mad_stream *stream)
{
  struct buffer *buffer = data;
  UINT unRead = 0, MaxRead;
  UINT RemainingBytes = stream->bufend - stream->next_frame;

  if (palReadPad(GPIOA,GPIOA_BUTTON))
  {
	  do
	  {
		  chThdSleepMilliseconds(100);
	  }
	  while (palReadPad(GPIOA,GPIOA_BUTTON));

	  return MAD_FLOW_STOP;
  }

  if ((buffer->eof)||(palReadPad(GPIOA,GPIOA_BUTTON)))
    return MAD_FLOW_STOP;
 
  palSetPad(GPIOD, GPIOD_LED5);

  if ((RemainingBytes)&&(RemainingBytes<MP3BUF))
    memmove(buffer->start, stream->next_frame, RemainingBytes);

  MaxRead = (MP3BUF - RemainingBytes)&0xFE00;

  FRESULT fr = f_read(buffer->file, buffer->start + RemainingBytes, MaxRead, &unRead);
  if (unRead!=MaxRead) buffer->eof = 1;

  //dbgprintf("Read : %d (rem=%d - MaxRead=%d)\n\r",unRead,RemainingBytes,MaxRead);

  mad_stream_buffer(stream, buffer->start, RemainingBytes+unRead);

  palClearPad(GPIOD, GPIOD_LED5);

  return MAD_FLOW_CONTINUE;
}

enum mad_flow header(void *data,
		    struct mad_header const *header)
{
  struct buffer *buffer = data;

 if (buffer->samplerate!=header->samplerate)
 {
  chprintf((BaseSequentialStream *)&SD,"MP3 Header Info :\n\r");
  chprintf((BaseSequentialStream *)&SD," Bitrate : %d\n\r",header->bitrate);
  chprintf((BaseSequentialStream *)&SD," Samplerate : %d\n\r",header->samplerate);

  Codec_AudioInterface_Init(header->samplerate);

  buffer->samplerate=header->samplerate;
 }

 return MAD_FLOW_CONTINUE;
}

/*
 * The following utility routine performs simple rounding, clipping, and
 * scaling of MAD's high-resolution samples down to 16 bits. It does not
 * perform any dithering or noise shaping, which would be recommended to
 * obtain any exceptional audio quality. It is therefore not recommended to
 * use this routine if high-quality output is desired.
 */

signed int scale(mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));
#if 0
  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
#else
  { signed int result;
    asm("ssat %0, %2, %1, asr %3" \
        : "=r" (result) \
        : "%r" (sample), \
          "M" (16), "M" (MAD_F_FRACBITS + 1 - 16) );
    return result; }
#endif
}

/*
 * This is the output callback function. It is called after each frame of
 * MPEG audio data has been completely decoded. The purpose of this callback
 * is to output (or play) the decoded PCM audio.
 */

#define OFFSET 0

enum mad_flow output(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
  unsigned int nchannels, nsamples;
  mad_fixed_t const *left_ch, *right_ch;
  signed sample;
  uint16_t *output=wave_buf1;
  uint16_t *write_buf;
  int i;
 // UINT unWrite;

  /* pcm->samplerate contains the sampling frequency */

  nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  palClearPad(GPIOD, GPIOD_LED6);
  if (playing!=0) chEvtWaitAny((eventmask_t)1);
  palSetPad(GPIOD, GPIOD_LED6);

  if (SPID3.dmatx->stream->CR & 0x80000) output=wave_buf0;

  write_buf=output;

  while (nsamples--) {
    sample = scale(*left_ch++);
    *output++=sample;
    if (nchannels == 2) sample = scale(*right_ch++);
    *output++=sample;
  }

  //FRESULT fr = f_write(&fOut, write_buf, 4*pcm->length, &unWrite);

  if (playing==0) Audio_MAL_Play(wave_buf0, wave_buf1, 4*1152);
  playing=1;

  return MAD_FLOW_CONTINUE;
}

/*
 * This is the error callback function. It is called whenever a decoding
 * error occurs. The error is indicated by stream->error; the list of
 * possible MAD_ERROR_* errors can be found in the mad.h (or stream.h)
 * header file.
 */

enum mad_flow error(void *data,
		    struct mad_stream *stream,
		    struct mad_frame *frame)
{
  struct buffer *buffer = data;

  dbgprintf("decoding error 0x%04x (%s) at byte offset %u\r\n",
	  stream->error, mad_stream_errorstr(stream),
	  stream->this_frame - buffer->start);

  /* return MAD_FLOW_BREAK here to stop decoding (and propagate an error) */

  return MAD_FLOW_CONTINUE;
}

/*
 * This is the function called by main() above to perform all the decoding.
 * It instantiates a decoder object and configures it with the input,
 * output, and error callback functions above. A single call to
 * mad_decoder_run() continues until a callback function returns
 * MAD_FLOW_STOP (to stop decoding) or MAD_FLOW_BREAK (to stop decoding and
 * signal an error).
 */


int decode(char *file)
{
  int result,i;

  FRESULT errFS = f_open(&fIn, file, FA_READ);
  if(errFS != FR_OK)
  {
    chprintf((BaseSequentialStream *)&SD, "Mp3Decode: Failed to open file \"%s\" for reading, err=%d\r\n", file, errFS);
    return -1;
  }

  //file[strlen(file)-1]='O';
  //errFS = f_open(&fOut, file, FA_WRITE|FA_CREATE_ALWAYS);
  //if(errFS != FR_OK)
  //{
  //  chprintf((BaseChannel*)&SD, "Mp3Decode: Failed to open file \"%s\" for writing, err=%d\r\n", file, errFS);
  //  return -1;
  //}

  /* initialize our private message structure */
  buffer.start  = mp3buf;
  buffer.file   = &fIn;
  buffer.eof    = 0;
  buffer.samplerate = 0;

  /* configure input, output, and error functions */
  mad_decoder_init(&decoder, &buffer,
		   input, header, 0 /* filter */, output,
		   error, 0 /* message */);

  playing=0;

  /* start decoding */
  result = mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);

  /* release the decoder */
  mad_decoder_finish(&decoder);
  f_close(&fIn);
  //f_close(&fOut);

  dbgprintf("Decoding return %d\r\n",result);

  return result;
}
