/*
    ChibiOS/RT - Copyright (C) 2006,2007,2008,2009,2010,
                 2011,2012 Giovanni Di Sirio.

    This file is part of ChibiOS/RT.

    ChibiOS/RT is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    ChibiOS/RT is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

                                      ---

    A special exception to the GPL can be applied should you wish to distribute
    a combined work that includes ChibiOS/RT, without being obliged to provide
    the source code for any proprietary components. See the file exception.txt
    for full details of how and when the exception can be applied.
*/

#include "ch.h"
#include "hal.h"
#include "test.h"

#include "chprintf.h"
#include "shell.h"

#include "evtimer.h"
#include <ff.h>
#include <string.h>

#include "mad.h"

static VirtualTimer tmr;
#define POLLING_DELAY 10
#define POLLING_INTERVAL 10
static unsigned tmr_cnt;

static EventSource inserted_event, removed_event;

FATFS MMC_FS;
MMCDriver MMCD1;
static bool_t fs_ready = FALSE;

#define SD SDU1

/* SPIConfig => callback(or NULL), port, pad, CR1 config */
static SPIConfig hs_spicfg = { NULL, GPIOC, 4, 0 };
static SPIConfig ls_spicfg = { NULL, GPIOC, 4, SPI_CR1_BR_2 | SPI_CR1_BR_1 };
// Was : mmcObjectInit(&MMCD1, &SPID1, &ls_spicfg, &hs_spicfg, mmc_is_protected, mmc_is_inserted);
MMCConfig MMCD1_Config={ &SPID1, &ls_spicfg, &hs_spicfg} ;

Thread *MadThd;

#if 0
bool_t mmc_lld_is_write_protected(MMCDriver *sdcp) {
  (void)sdcp;
  return FALSE;
}

bool_t mmc_lld_is_card_inserted(MMCDriver *sdcp) {
  (void)sdcp;
  // not wired yet, assume card is always in
  //return palReadPad(IOPORT3, GPIOC_MMCCP);
  return 1;
}
#endif

static void tmrfunc(void *p) {
 chSysLockFromIsr();

 if (tmr_cnt>0) {
  if (mmc_lld_is_card_inserted(NULL)) { if (--tmr_cnt == 0) {chEvtBroadcastI(&inserted_event);}}
                                else  { tmr_cnt = POLLING_INTERVAL; }
 }
 else {
  if (!mmc_lld_is_card_inserted(NULL)) { tmr_cnt = POLLING_INTERVAL; chEvtBroadcastI(&removed_event); }
 }
                                              
 chVTSetI(&tmr, MS2ST(POLLING_DELAY), tmrfunc, NULL);
 chSysUnlockFromIsr();
}

static void tmr_init(void *p) {
 chEvtInit(&inserted_event);
 chEvtInit(&removed_event);
 chSysLock();
 tmr_cnt=POLLING_INTERVAL;
 chVTSetI(&tmr, MS2ST(POLLING_DELAY), tmrfunc, p);
 chSysUnlock();
}

#include "usb-shell.c"

#include "mad.c"

FRESULT Mp3PlayAllFiles(char* path)
{
  FILINFO fno;
  DIR dir;
  char fpath[50];

  chprintf((BaseSequentialStream*)&SD, "Mp3PlayAllFiles: Playing all files in \"%s\"\r\n", path);
  FRESULT res = f_opendir(&dir, path);
  if(res == FR_OK)
  {
    res = f_readdir(&dir, &fno);
    while((res == FR_OK) && (fno.fname[0] != 0))
    {
      strcpy(fpath, path);
      strcat(fpath, fno.fname);
      if(fno.fattrib & AM_DIR)
      {
        chprintf((BaseSequentialStream*)&SD, "Mp3PlayAllFiles: DIR \"%s\"\r\n", fpath);
      }
      else
      {
        chprintf((BaseSequentialStream*)&SD, "Mp3PlayAllFiles: FILE \"%s\"\r\n", fpath);

        size_t unLen = strlen(fpath);
        if(unLen > 3)
        {
          if(stricmp(fpath + unLen - 4, ".mp3") == 0)
          {
	    decode(fpath);
          }
        }
      }
      res = f_readdir(&dir, &fno);
    }
  }
  else
  {
    chprintf((BaseSequentialStream*)&SD, "Mp3PlayAllFiles: f_opendir failed %d\r\n", res);
  }
  chprintf((BaseSequentialStream*)&SD, "Mp3PlayAllFiles: Finished playing all files\r\n", path);

  return res;
}

// MMC card insertion event
static void InsertHandler(eventid_t id)
{
  FRESULT err;
  (void) id;
  if(mmcConnect(&MMCD1))
  {
    chprintf((BaseSequentialStream *) &SD, "SD: Failed to connect to card\r\n");
    return;
  }
  else
  {
    chprintf((BaseSequentialStream *) &SD, "SD: Connected to card\r\n");
  }

  err = f_mount(0, &MMC_FS);
  if(err != FR_OK)
  {
    chprintf((BaseSequentialStream *) &SD, "SD: f_mount() failed %d\r\n", err);
    mmcDisconnect(&MMCD1);
    return;
  }
  else
  {
    chprintf((BaseSequentialStream *) &SD, "SD: File system mounted\r\n");
  }
  fs_ready = TRUE;

  // quick and dirty
  // directly decode all mp3's in the
  // cards root in a loop (assuming the card is never pulled)
  MadThd=chThdSelf();
  while(1)
  {
    Mp3PlayAllFiles("/");
  }
}

// MMC card removal event
static void RemoveHandler(eventid_t id)
{
  (void) id;
  fs_ready = FALSE;
}

// called on kernel panic
void port_halt(void)
{
  port_disable();
  while(TRUE)
  {
    palSetPad(GPIOD, GPIOD_LED4);       /* Green */
  }
}

void debug(void)
{
  chThdSleepMilliseconds(1000);

  chprintf((BaseSequentialStream *)&SD, "I2S:\r\n");
  chprintf((BaseSequentialStream *)&SD, " CR1 %x\r\n", SPI3->CR1);
  chprintf((BaseSequentialStream *)&SD, " CR2 %x\r\n", SPI3->CR2);
  chprintf((BaseSequentialStream *)&SD, " SR  %x\r\n", SPI3->SR);
  chprintf((BaseSequentialStream *)&SD, " CFG %x\r\n", SPI3->I2SCFGR);
  chprintf((BaseSequentialStream *)&SD, " PR  %x\r\n", SPI3->I2SPR);

  chprintf((BaseSequentialStream *)&SD, "DMA:\r\n");
  chprintf((BaseSequentialStream *)&SD, " CR  %x\r\n", SPID3.dmatx->stream->CR);
  chprintf((BaseSequentialStream *)&SD, " NDT %x\r\n", SPID3.dmatx->stream->NDTR);
  chprintf((BaseSequentialStream *)&SD, " PAR %x\r\n", SPID3.dmatx->stream->PAR);
  chprintf((BaseSequentialStream *)&SD, " M0A %x\r\n", SPID3.dmatx->stream->M0AR);
  chprintf((BaseSequentialStream *)&SD, " M1A %x\r\n", SPID3.dmatx->stream->M1AR);
  chprintf((BaseSequentialStream *)&SD, " FCR %x\r\n", SPID3.dmatx->stream->FCR);
}

/*
 * This is a periodic thread that does absolutely nothing except flashing
 * a LED.
 */
static WORKING_AREA(waThread1, 128);
__attribute__((noreturn)) static msg_t Thread1(void *arg) {

  (void)arg;
  chRegSetThreadName("blinker");
  while (TRUE) {
    palSetPad(GPIOD, GPIOD_LED3);       /* Orange.  */
    chThdSleepMilliseconds(500);
    palClearPad(GPIOD, GPIOD_LED3);     /* Orange.  */
    chThdSleepMilliseconds(500);
    //debug();
  }
}

uint32_t EVAL_AUDIO_Init(uint16_t OutputDevice, uint8_t Volume, uint32_t AudioFreq);

/*
 * Application entry point.
 */
int main(void) {
  static Thread *shelltp = NULL;
  static const evhandler_t evhndl[] = {InsertHandler, RemoveHandler};
  struct EventListener el0, el1;

  /*
   * System initializations.
   * - HAL initialization, this also initializes the configured device drivers
   *   and performs the board-specific initializations.
   * - Kernel initialization, the main() function becomes a thread and the
   *   RTOS is active.
   */
  halInit();
  chSysInit();

#if 0
  /*
   * Activates the serial driver 2 using the driver default configuration.
   * PA2(TX) and PA3(RX) are routed to USART2.
   */
  sdStart(&SD, NULL);
  palSetPadMode(GPIOA, 2, PAL_MODE_ALTERNATE(7));
  palSetPadMode(GPIOA, 3, PAL_MODE_ALTERNATE(7));
#else
  sduObjectInit(&SDU1);
  sduStart(&SDU1, &serusbcfg);

  /*
   * Activates the USB driver and then the USB bus pull-up on D+.
   * Note, a delay is inserted in order to not have to disconnect the cable
   * after a reset.
   */
  usbDisconnectBus(serusbcfg.usbp);
  chThdSleepMilliseconds(1500);
  usbStart(serusbcfg.usbp, &usbcfg);
  usbConnectBus(serusbcfg.usbp);

  /*
   * Shell manager initialization.
   */
  shellInit();

  /*
   * Activates the serial driver 6 and SDC driver 1 using default
   * configuration.
   */
  sdStart(&SD6, NULL);
#endif

  // setup pads to SPI1 function (connect these pads to your SD card accordingly)
  palSetPadMode(GPIOC, 4, PAL_MODE_OUTPUT_PUSHPULL | PAL_STM32_OSPEED_HIGHEST); // NSS
  palSetPadMode(GPIOA, 5, PAL_MODE_ALTERNATE(5) | PAL_STM32_OSPEED_HIGHEST); // SCK
  palSetPadMode(GPIOA, 6, PAL_MODE_ALTERNATE(5)); // MISO
  palSetPadMode(GPIOA, 7, PAL_MODE_ALTERNATE(5) | PAL_STM32_OSPEED_HIGHEST); // MOSI
  palSetPad(GPIOC, 4); // set NSS high

  // initialize MMC driver
  mmcObjectInit(&MMCD1);
  mmcStart(&MMCD1, &MMCD1_Config);

  EVAL_AUDIO_Init(0,70,44100);

  /*
   * Creates the example thread.
   */
  chThdCreateStatic(waThread1, sizeof(waThread1), NORMALPRIO, Thread1, NULL);


  chprintf((BaseSequentialStream *)&SD, "STM32 Audio MP3 / ChibiOS / Libmad : %s %s %s\r\n", __FILE__, __DATE__, __TIME__);
  chprintf((BaseSequentialStream *)&SD, "%s\r\n", mad_version);
  chprintf((BaseSequentialStream *)&SD, "%s\r\n", mad_copyright);
  chprintf((BaseSequentialStream *)&SD, "%s\r\n", mad_author);
  chprintf((BaseSequentialStream *)&SD, "%s\r\n", mad_build);

  // GO

  tmr_init(NULL);

  chEvtRegister(&inserted_event, &el0, 0);
  chEvtRegister(&removed_event, &el1, 1);
  while(TRUE)
  {

    if (!shelltp && (SDU1.config->usbp->state == USB_ACTIVE))
      shelltp = shellCreate(&shell_cfg1, SHELL_WA_SIZE, NORMALPRIO);
    else if (chThdTerminated(shelltp)) {
      chThdRelease(shelltp);    /* Recovers memory of the previous shell.   */
      shelltp = NULL;           /* Triggers spawning of a new shell.        */
    }
 
    chEvtDispatch(evhndl, chEvtWaitOneTimeout(ALL_EVENTS , MS2ST(500)));
  }
}
